@@ -46,11 +46,11 @@ typedef long Py_hash_t;
 #endif
 
 #if PY_MAJOR_VERSION >= 3
-#define PY_STRING(x) PyUnicode_FromString(x)
+#define PY_STRING(x) PyUnicode_DecodeUTF8(x, strlen(x), "ignore" )
 #define PY_STRING_TO_C(x) PyUnicode_AsUTF8(x)
 #define PY_STRING_CHECK(x) PyUnicode_Check(x)
 #else
-#define PY_STRING(x) PyString_FromString(x)
+#define PY_STRING(x) PyString_Decode(x, strlen(x), "utf-8", "ignore")
 #define PY_STRING_TO_C(x) PyString_AsString(x)
 #define PY_STRING_CHECK(x) (PyString_Check(x) || PyUnicode_Check(x))
 #endif
@@ -121,7 +121,7 @@ static PyMemberDef Match_members[] = {
     T_OBJECT_EX,
     offsetof(Match, strings),
     READONLY,
-    "Dictionary with offsets and strings that matched the file"
+    "Tuple with offsets and strings that matched the file"
   },
   { NULL } // End marker
 };
@@ -206,6 +206,8 @@ typedef struct
   PyObject* identifier;
   PyObject* tags;
   PyObject* meta;
+  PyObject* global;
+  PyObject* private;
 } Rule;
 
 static void Rule_dealloc(
@@ -216,6 +218,20 @@ static PyObject* Rule_getattro(
     PyObject* name);
 
 static PyMemberDef Rule_members[] = {
+  {
+    "is_global",
+    T_OBJECT_EX,
+    offsetof(Rule, global),
+    READONLY,
+    "Rule is global"
+  },
+  {
+    "is_private",
+    T_OBJECT_EX,
+    offsetof(Rule, private),
+    READONLY,
+    "Rule is private"
+  },
   {
     "identifier",
     T_OBJECT_EX,
@@ -423,7 +439,7 @@ PyObject* convert_object_to_python(
   switch(object->type)
   {
     case OBJECT_TYPE_INTEGER:
-      if (object->value.i != UNDEFINED)
+      if (object->value.i != YR_UNDEFINED)
         result = Py_BuildValue("l", object->value.i);
       break;
 
@@ -508,7 +524,7 @@ PyObject* convert_array_to_python(
   if (array->items == NULL)
     return py_list;
 
-  for (i = 0; i < array->items->count; i++)
+  for (i = 0; i < array->items->length; i++)
   {
     py_object = convert_object_to_python(array->items->objects[i]);
 
@@ -546,7 +562,7 @@ PyObject* convert_dictionary_to_python(
     {
       PyDict_SetItemString(
           py_dict,
-          dictionary->items->objects[i].key,
+          dictionary->items->objects[i].key->c_string,
           py_object);
 
       Py_DECREF(py_object);
@@ -562,6 +578,7 @@ PyObject* convert_dictionary_to_python(
 #define CALLBACK_ALL CALLBACK_MATCHES | CALLBACK_NON_MATCHES
 
 int yara_callback(
+    YR_SCAN_CONTEXT* context,
     int message,
     void* message_data,
     void* user_data)
@@ -724,7 +741,7 @@ int yara_callback(
 
   yr_rule_strings_foreach(rule, string)
   {
-    yr_string_matches_foreach(string, m)
+    yr_string_matches_foreach(context, string, m)
     {
       object = PyBytes_FromStringAndSize((char*) m->data, m->data_length);
 
@@ -991,7 +1008,7 @@ int process_compile_externals(
       result = yr_compiler_define_integer_variable(
           compiler,
           identifier,
-          PyLong_AsLong(value));
+          PyLong_AsLongLong(value));
     }
     else if (PyFloat_Check(value))
     {
@@ -1061,7 +1078,7 @@ int process_match_externals(
       result = yr_rules_define_integer_variable(
           rules,
           identifier,
-          PyLong_AsLong(value));
+          PyLong_AsLongLong(value));
     }
     else if (PyFloat_Check(value))
     {
@@ -1239,6 +1256,8 @@ static void Rule_dealloc(
   Py_XDECREF(object->identifier);
   Py_XDECREF(object->tags);
   Py_XDECREF(object->meta);
+  Py_XDECREF(object->global);
+  Py_XDECREF(object->private);
   PyObject_Del(self);
 }
 
@@ -1325,6 +1344,8 @@ static PyObject* Rules_next(
       Py_DECREF(object);
     }
 
+    rule->global = PyBool_FromLong(rules->iter_current_rule->flags & RULE_FLAGS_GLOBAL);
+    rule->private = PyBool_FromLong(rules->iter_current_rule->flags & RULE_FLAGS_PRIVATE);
     rule->identifier = PY_STRING(rules->iter_current_rule->identifier);
     rule->tags = tag_list;
     rule->meta = meta_list;
@@ -1455,7 +1476,6 @@ static PyObject* Rules_match(
       callback_data.matches = PyList_New(0);
 
       Py_BEGIN_ALLOW_THREADS
-
       error = yr_rules_scan_file(
           object->rules,
           filepath,
@@ -1464,6 +1484,7 @@ static PyObject* Rules_match(
           &callback_data,
           timeout);
 
+
       Py_END_ALLOW_THREADS
     }
     else if (data != NULL)
@@ -1664,6 +1685,7 @@ void raise_exception_on_error(
     int error_level,
     const char* file_name,
     int line_number,
+    const YR_RULE* rule,
     const char* message,
     void* user_data)
 {
@@ -1690,6 +1712,7 @@ void raise_exception_on_error_or_warning(
     int error_level,
     const char* file_name,
     int line_number,
+    const YR_RULE* rule,
     const char* message,
     void* user_data)
 {
@@ -1837,20 +1860,22 @@ static PyObject* yara_set_config(
    * options present in yara/libyara.c yr_set_configuration(...) - ck
    */
   static char *kwlist[] = {
-    "stack_size", "max_strings_per_rule", NULL};
+    "stack_size", "max_strings_per_rule", "max_match_data", NULL};
 
   unsigned int stack_size = 0;
   unsigned int max_strings_per_rule = 0;
+  unsigned int max_match_data = 0;
 
   int error = 0;
 
   if (PyArg_ParseTupleAndKeywords(
         args,
         keywords,
-        "|II",
+        "|III",
         kwlist,
         &stack_size,
-        &max_strings_per_rule))
+        &max_strings_per_rule,
+  	&max_match_data))
   {
     if (stack_size != 0)
     {
@@ -1871,6 +1896,16 @@ static PyObject* yara_set_config(
       if (error != ERROR_SUCCESS)
         return handle_error(error, NULL);
     }
+
+    if (max_match_data != 0)
+    {
+      error = yr_set_configuration(
+          YR_CONFIG_MAX_MATCH_DATA,
+				  &max_match_data);
+
+      if (error != ERROR_SUCCESS)
+        return handle_error(error, NULL);
+    }
   }
 
   Py_RETURN_NONE;
@@ -2309,7 +2344,7 @@ static PyMethodDef yara_methods[] = {
     "set_config",
     (PyCFunction) yara_set_config,
     METH_VARARGS | METH_KEYWORDS,
-    "Set a yara configuration variable (stack_size or max_strings_per_rule)"
+    "Set a yara configuration variable (stack_size, max_strings_per_rule, or max_match_data)"
   },
   { NULL, NULL }
 };
