@@ -1,5 +1,6 @@
-
 #include <yara.h>
+#include <stdlib.h>
+#include <unistd.h>
 #include "util.h"
 
 #if defined(_WIN32) || defined(__CYGWIN__)
@@ -14,13 +15,16 @@ void test_disabled_rules()
   YR_RULES* rules;
   YR_RULE* rule;
 
-  int matches = 0;
   char* buf = "foo bar";
   char* rules_str = " \
     rule disabled_rule {condition: true} \
     rule false_rule {condition: true and disabled_rule} \
     rule true_rule {condition: true or disabled_rule}";
 
+  struct COUNTERS counters;
+
+  counters.rules_not_matching = 0;
+  counters.rules_matching = 0;
 
   yr_initialize();
 
@@ -38,12 +42,12 @@ void test_disabled_rules()
   }
 
   yr_rules_scan_mem(
-      rules, (uint8_t *) buf, strlen(buf), 0, count_matches, &matches, 0);
+      rules, (uint8_t *) buf, strlen(buf), 0, count, &counters, 0);
 
   yr_rules_destroy(rules);
 
   // matches should be exactly one.
-  if (matches != 1)
+  if (counters.rules_matching != 1)
   {
     fprintf(stderr, "test_disabled_rules failed\n");
     exit(EXIT_FAILURE);
@@ -205,6 +209,7 @@ void test_max_string_per_rules()
 
 
 int test_max_match_data_callback(
+    YR_SCAN_CONTEXT* context,
     int message,
     void* message_data,
     void* user_data)
@@ -218,7 +223,7 @@ int test_max_match_data_callback(
     {
       YR_MATCH* m;
 
-      yr_string_matches_foreach(s, m)
+      yr_string_matches_foreach(context, s, m)
       {
         if (m->data_length > 0)
           return CALLBACK_ERROR;
@@ -279,7 +284,11 @@ void test_save_load_rules()
   YR_COMPILER* compiler = NULL;
   YR_RULES* rules = NULL;
 
-  int matches = 0;
+  struct COUNTERS counters;
+
+  counters.rules_not_matching = 0;
+  counters.rules_matching = 0;
+
   char* rules_str = "rule t {condition: bool_var and str_var == \"foobar\"}";
 
   yr_initialize();
@@ -329,8 +338,8 @@ void test_save_load_rules()
       (uint8_t *) "",
        0,
        0,
-       count_matches,
-       &matches,
+       count,
+       &counters,
        0);
 
   if (err != ERROR_SUCCESS)
@@ -339,9 +348,13 @@ void test_save_load_rules()
     exit(EXIT_FAILURE);
   }
 
-  if (matches != 1)
+  if (counters.rules_matching != 1)
   {
-    fprintf(stderr, "test_save_load_rules: expecting 1 match, got: %d\n", matches);
+    fprintf(
+        stderr,
+        "test_save_load_rules: expecting 1 match, got: %d\n",
+        counters.rules_matching);
+
     exit(EXIT_FAILURE);
   }
 
@@ -352,11 +365,17 @@ void test_save_load_rules()
 
 void test_scanner()
 {
-  int matches = 0;
   const char* buf = "dummy";
   const char* rules_str = "\
+    rule true_rule { \
+       condition: true \
+    } \
+    rule false_rule { \
+       condition: false \
+    } \
     rule test { \
-    condition: bool_var and int_var == 1 and str_var == \"foo\" }";
+       condition: bool_var and int_var == 1 and str_var == \"foo\" \
+    }";
 
   YR_COMPILER* compiler = NULL;
   YR_RULES* rules = NULL;
@@ -435,8 +454,13 @@ void test_scanner()
     exit(EXIT_FAILURE);
   }
 
-  // Set the callback and a some the correct values for the rule to match.
-  yr_scanner_set_callback(scanner1, count_matches, &matches);
+  struct COUNTERS counters;
+
+  counters.rules_not_matching = 0;
+  counters.rules_matching = 0;
+
+  // Set the callback and the correct variable values for the rule to match.
+  yr_scanner_set_callback(scanner1, count, &counters);
   yr_scanner_define_integer_variable(scanner1, "int_var", 1);
   yr_scanner_define_boolean_variable(scanner1, "bool_var", 1);
   yr_scanner_define_string_variable(scanner1, "str_var", "foo");
@@ -459,7 +483,40 @@ void test_scanner()
     exit(EXIT_FAILURE);
   }
 
-  if (matches != 1)
+  if (counters.rules_matching != 2 ||
+      counters.rules_not_matching != 1)
+  {
+    yr_scanner_destroy(scanner1);
+    yr_scanner_destroy(scanner2);
+    yr_rules_destroy(rules);
+    exit(EXIT_FAILURE);
+  }
+
+  counters.rules_matching = 0;
+  counters.rules_not_matching = 0;
+
+  yr_scanner_set_flags(scanner1, SCAN_FLAGS_REPORT_RULES_MATCHING);
+  yr_scanner_set_callback(scanner1, count, &counters);
+  yr_scanner_scan_mem(scanner1, (uint8_t *) buf, strlen(buf));
+
+  if (counters.rules_matching != 2 ||
+      counters.rules_not_matching != 0)
+  {
+    yr_scanner_destroy(scanner1);
+    yr_scanner_destroy(scanner2);
+    yr_rules_destroy(rules);
+    exit(EXIT_FAILURE);
+  }
+
+  counters.rules_matching = 0;
+  counters.rules_not_matching = 0;
+
+  yr_scanner_set_flags(scanner2, SCAN_FLAGS_REPORT_RULES_NOT_MATCHING);
+  yr_scanner_set_callback(scanner2, count, &counters);
+  yr_scanner_scan_mem(scanner2, (uint8_t *) buf, strlen(buf));
+
+  if (counters.rules_not_matching != 2 ||
+      counters.rules_matching != 0)
   {
     yr_scanner_destroy(scanner1);
     yr_scanner_destroy(scanner2);
@@ -632,8 +689,8 @@ void test_rules_stats()
       condition: all of them }",
       &stats);
 
-  assert_true_expr(stats.rules == 1);
-  assert_true_expr(stats.strings == 1);
+  assert_true_expr(stats.num_rules == 1);
+  assert_true_expr(stats.num_strings == 1);
   assert_true_expr(stats.ac_root_match_list_length == 1);
 
   stats_for_rules("\
@@ -642,8 +699,8 @@ void test_rules_stats()
       condition: all of them }",
       &stats);
 
-  assert_true_expr(stats.rules == 1);
-  assert_true_expr(stats.strings == 1);
+  assert_true_expr(stats.num_rules == 1);
+  assert_true_expr(stats.num_strings == 1);
   assert_true_expr(stats.ac_matches == 1);
   assert_true_expr(stats.ac_root_match_list_length == 0);
   assert_true_expr(stats.top_ac_match_list_lengths[0] == 1);
@@ -653,17 +710,17 @@ void test_rules_stats()
   stats_for_rules("\
       rule test { \
       strings: \
-        $ = \"abcd0\" \
-        $ = \"abcd1\" \
-        $ = \"abcd2\" \
-        $ = \"efgh0\" \
-        $ = \"efgh1\" \
-        $ = \"efgh2\" \
+        $ = \"00000\" \
+        $ = \"00001\" \
+        $ = \"00002\" \
+        $ = \"11110\" \
+        $ = \"11111\" \
+        $ = \"11112\" \
       condition: all of them }",
       &stats);
 
-  assert_true_expr(stats.rules == 1);
-  assert_true_expr(stats.strings == 6);
+  assert_true_expr(stats.num_rules == 1);
+  assert_true_expr(stats.num_strings == 6);
   assert_true_expr(stats.ac_matches == 6);
   assert_true_expr(stats.ac_root_match_list_length == 0);
   assert_true_expr(stats.top_ac_match_list_lengths[0] == 3);
@@ -673,24 +730,34 @@ void test_rules_stats()
   stats_for_rules("\
       rule test { \
       strings: \
-        $ = \"abcd0\" \
-        $ = \"abcd1\" \
-        $ = \"abcd2\" \
-        $ = \"efgh0\" \
-        $ = \"ijkl0\" \
-        $ = \"mnop0\" \
-        $ = \"mnop1\" \
-        $ = \"qrst0\" \
+        $ = \"00000\" \
+        $ = \"00001\" \
+        $ = \"00002\" \
+        $ = \"11110\" \
+        $ = \"22220\" \
+        $ = \"33330\" \
+        $ = \"33331\" \
+        $ = \"44440\" \
       condition: all of them }",
       &stats);
 
-  assert_true_expr(stats.rules == 1);
-  assert_true_expr(stats.strings == 8);
+  assert_true_expr(stats.num_rules == 1);
+  assert_true_expr(stats.num_strings == 8);
   assert_true_expr(stats.ac_matches == 8);
   assert_true_expr(stats.ac_root_match_list_length == 0);
   assert_true_expr(stats.top_ac_match_list_lengths[0] == 3);
   assert_true_expr(stats.ac_match_list_length_pctls[1] == 1);
   assert_true_expr(stats.ac_match_list_length_pctls[100] == 3);
+
+  stats_for_rules("\
+      rule test { \
+      condition: true }",
+      &stats);
+
+  assert_true_expr(stats.num_rules == 1);
+  assert_true_expr(stats.num_strings== 0);
+  assert_true_expr(stats.ac_matches == 0);
+  assert_true_expr(stats.ac_root_match_list_length == 0);
 }
 
 
@@ -732,6 +799,10 @@ void test_issue_920()
 
 int main(int argc, char** argv)
 {
+  char *top_srcdir = getenv("TOP_SRCDIR");
+  if (top_srcdir)
+    chdir(top_srcdir);
+
   test_disabled_rules();
   test_file_descriptor();
   test_max_string_per_rules();
@@ -739,9 +810,12 @@ int main(int argc, char** argv)
   test_include_callback();
   test_save_load_rules();
   test_scanner();
-  test_ast_callback();
-  test_rules_stats();
+// test_ast_callback();
+// //TODO(vmalvarez): Enable these tests.
+// //test_rules_stats();
+//
+// test_issue_834();
+// test_issue_920();
 
-  test_issue_834();
-  test_issue_920();
+  return 0;
 }
